// SkipList.java
//
// ICS 23 / CSE 23 Summer 2012
// Project #3: Always Changing Probably
//
// Below is a skeleton of a SkipList class where each key is an int and each
// value is an ArrayList<Integer>.  This data structure can be used to
// implement the database tables in this project, with each key/value pair
// being one row in the table, where the key is the element stored in the
// key column and the value is an ArrayList of Integers that are the values
// in each of the named columns.
//
// The skeleton suggests a basic design for your class, which you're free to
// use or discard at your discretion, though I suggest you stick with it, as
// I've already successfully implemented the SkipList according to this design
// and found it to work well.

import java.util.ArrayList;



public class SkipList
{
	// The constructor initializes a newly-constructed SkipList to be empty.
	// Remember that an empty skip list actually contains one level, numbered
	// "level 0," that contains only the special -INF and +INF nodes.
	public SkipList()
	{
	}
	

	// insert() takes a key/value pair as parameters, adding the key and its
	// associated value into the skip list.  If the key already exists in the
	// skip list, a DuplicateKeyException is thrown.	
	public void insert(int key, ArrayList<Integer> value)
	throws DuplicateKeyException
	{
	}
	

	// lookup() takes a key as a parameter and returns the value associated
	// with that key in the skip list.  If the key does not appear in the skip
	// list, a NoSuchKeyException is thrown.	
	public ArrayList<Integer> lookup(int key)
	throws NoSuchKeyException
	{
	}
	

	// remove() takes a key as a parameter and removes it, along with its
	// associated value, from the skip list.  If the key does not appear in
	// the skip list, a NoSuchKeyException is thrown.	
	public void remove(int key)
	throws NoSuchKeyException
	{
	}
	

	// update() takes a key and a new value as parameters.  If the key appears
	// in the skip list already, its value is replaced with the new value; if
	// not, a NoSuchKeyException is thrown instead.	
	public void update(int key, ArrayList<Integer> newValue)
	throws NoSuchKeyException
	{
	}
	

	// There are three kinds of nodes in a skip list.  All kinds of nodes
	// have references that point to the node before, after, above, and
	// below that node.  But, other than that, there are differences between
	// the three kinds of nodes:
	//
	//   * Normal nodes contain a key/value pair.  Comparing a search key
	//     to a normal node is done by comparing the search key to the key
	//     in the node, so that a node containing the key 50 will be
	//     considered to come before the search key 60, while a node
	//     containing the key 70 will be considered to come after the
	//     search key 40.
	//
	//   * -INF nodes do not contain a key/value pair, so it is impossible
	//     to ask such a node for its key or its value.  Comparing a search
	//     key to a -INF node always has the same result: a -INF node is
	//     considered to come before any search key.
	//
	//   * +INF nodes are somewhat like -INF nodes, in that they contain no
	//     key/value pair.  However, when comparing a search key to a +INF
	//     node, the +INF node is always considered to come after any search
	//     key.
	//
	// I've provided a hierarchy of node classes that address these issues.
	// The abstract class Node forms the basis for all kinds of nodes,
	// containing the four references (prev, next, above, and below).  The
	// class NormalNode represents a normal node, with a key/value pair.
	// The abstract class InfinityNode forms the basis for the two kinds of
	// infinity nodes (-INF and +INF).  Finally, the classes NegativeInfinityNode
	// and PositiveInfinityNode represent -INF and +INF nodes, respectively.
	
	private abstract static class Node
	{
		public Node prev;
		public Node next;
		public Node above;
		public Node below;

		// Because nodes are interdependent, it's easier for all four references
		// to be set to null at construction time, then filled in one by one
		// afterward.
		public Node()
		{
			prev = null;
			next = null;
			above = null;
			below = null;
		}

		// getKey() returns the key stored in this node.  If the node is an
		// infinity node, an UnsupportedOperationException is thrown, since you
		// can't get a key out of an infinity node.
		public abstract int getKey();

		// getValue() returns the value stored in this node.  If the node is
		// an infinity node, an UnsupportedOperationException is thrown, since
		// you can't get a value out of an infinity node.
		public abstract ArrayList<Integer> getValue();

		// setValue() changes the value stored in this node.  If the node is
		// an infinity node, an UnsupportedOperationException is thrown, since
		// you can't get a value out of an infinity node.
		public abstract void setValue(ArrayList<Integer> value);

		// compareToSearchKey() compares the key in this node to the given
		// search key, returning a negative number if this key is less than
		// the search key, zero if they are equal, and a positive number if
		// this key is greater than the search key.  This method is handy
		// for implementing several parts of the skip list without having to
		// specially handle the cases of -INF and +INF (since polymorphism
		// will cause the appropriate version of compareToSearchKey() to be
		// called automatically).
		public abstract int compareToSearchKey(int searchKey);
	}


	private abstract static class InfinityNode
	extends Node
	{
		public InfinityNode()
		{
		}
		
		public int getKey()
		{
			throw new UnsupportedOperationException();
		}
		
		public ArrayList<Integer> getValue()
		{
			throw new UnsupportedOperationException();
		}
		
		public void setValue(ArrayList<Integer> value)
		{
			throw new UnsupportedOperationException();
		}
		
		public abstract int compareToSearchKey(int searchKey);
	}
	
	
	private static class NegativeInfinityNode
	extends InfinityNode
	{
		public NegativeInfinityNode()
		{
		}
		
		public int compareToSearchKey(int searchKey)
		{
			return -1;
		}
	}
	
	
	private static class PositiveInfinityNode
	extends InfinityNode
	{
		public PositiveInfinityNode()
		{
		}
		
		public int compareToSearchKey(int searchKey)
		{
			return 1;
		}
	}
	
	
	private static class NormalNode
	extends Node
	{
		private int key;
		private ArrayList<Integer> value;

		public NormalNode(int key, ArrayList<Integer> value)
		{
			this.key = key;
			this.value = value;
		}
				
		public int getKey()
		{
			return key;
		}
				
		public ArrayList<Integer> getValue()
		{
			return value;
		}
		
		public void setValue(ArrayList<Integer> value)
		{
			this.value = value;
		}
		
		public int compareToSearchKey(int searchKey)
		{
			if (key < searchKey)
			{
				return -1;
			}
			else if (key > searchKey)
			{
				return 1;
			}
			else
			{
				return 0;
			}
		}
	}
}
